/*

  Copyright (c) 1998, Digital Creations, Fredericksburg, VA, USA.  
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
  met:
  
    o Redistributions of source code must retain the above copyright
      notice, this list of conditions, and the disclaimer that follows.
  
    o Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions, and the following disclaimer in
      the documentation and/or other materials provided with the
      distribution.
  
    o Neither the name of Digital Creations nor the names of its
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.
  
  
  THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS AND CONTRIBUTORS *AS
  IS* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
  TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL
  CREATIONS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.

  If you have questions regarding this software,
  contact:
 
    Digital Creations L.C.  
    info@digicool.com
 
    (540) 371-6909

*/

typedef union {
  char c[1];
#define CS__DECLARE(T) CS_ ## T T[1];
  CS__DECLARE(CHAR);
  CS__DECLARE(DATETIME);
  CS__DECLARE(DATETIME4);
  CS__DECLARE(BIT);
  CS__DECLARE(TINYINT);
  CS__DECLARE(SMALLINT);
  CS__DECLARE(INT);
  CS__DECLARE(NUMERIC);
  CS__DECLARE(FLOAT);
  CS__DECLARE(REAL);
  CS__DECLARE(MONEY);
  CS__DECLARE(MONEY4);
} BufferUnion;

typedef struct {
  PyObject_VAR_HEAD
  CS_INT element_type;
  int element_size;
  BufferUnion val;
} Buffer;

#define STRING_CASES CS_CHAR_TYPE: case CS_LONGCHAR_TYPE: case CS_BINARY_TYPE: case CS_LONGBINARY_TYPE: case CS_TEXT_TYPE: case CS_IMAGE_TYPE


static PyTypeObject BufferType;

#define Buffer_Check(O) ((O)->ob_type == &BufferType)

static void
Buffer_dealloc(Buffer *self) { PyMem_DEL(self); }

static PyObject *
Buffer_str(Buffer *self)
{
  return PyString_FromStringAndSize(self->val.c, self->ob_size);
}

static int
Buffer_length(Buffer *self)
{
  return self->ob_size/self->element_size;
}

static PyObject * 
Buffer_concat(Buffer *self, PyObject *bb)
{
  PyErr_SetString(PyExc_TypeError,
		  "Buffer objects do not support concatenation");
  return NULL;
}

static PyObject *
Buffer_repeat(Buffer *self, int n)
{
  PyErr_SetString(PyExc_TypeError,
		  "Buffer objects do not support repetition");
  return NULL;
}

static PyObject *
index_error(int i)
{
  PyObject *r;

  if((r=PyInt_FromLong(i)))
    {
      PyErr_SetObject(PyExc_IndexError, r);
      Py_DECREF(r);
    }
  return NULL;
}

static int
year_as_day(int y)
{
  int t;

  y -= 1900;
  t = (y - 1) * 365;
  if(y > 0) y--;
  return t + y/4 - y/100 + (y+300)/400;
}
       
static int
year_of_day(int days)
{
  int yad, y;

  y = days/366 + 1901;
  yad = year_as_day(y);
  if(yad > days) y--;
  while(year_as_day(y) <= days) y++;
  y--;
  return y;
}

static int
length_of_month(int y, int m)
{
  static int month_length[] = {0, 31, 29, 31, 30, 31, 30, 31, 31, 30,
			       31, 30, 31};
  
  if(m != 2) return month_length[m];
  if(y%4 == 0 && (y%400 == 0 || y%100 != 0)) return 29;
  return 28;
}

static void
ymd(int days, int *y, int *m, int *d)
{
  int l;
  int year, month, day;

  days-=365;
  year = year_of_day(days);
  day = days - year_as_day(year);
  for(month=1; month<=12; month++)
    {
      l = length_of_month(year,month);
      if(l > day)
	break;
      day -= l;
    }

  *d=day+1;
  *y=year;
  *m=month;
}

static int
days(int year, int month, int day)
{
  static int daysUntilMonth[] = 
    {0,0,31,59,90,120,151,181,212,243,273,304,334};
  int ys1900;
  long days;

  ys1900 = year - 1900;
  days = day + daysUntilMonth[month] + 365 * (ys1900 - 1);
  if(ys1900 > 0)  ys1900--;
  days += ys1900 / 4 - ys1900 / 100 + (ys1900 + 300) / 400 - 1;
  if(month > 2 && 
     (year % 4 == 0 && (year % 100 != 0 || year % 400 == 0)))
    days++;
  return days;
}

static PyObject *
Buffer_item(Buffer *self, int i)
{
  int ii, s, es;
  PyObject *h=0, *l=0;
  int year, month, day, hour, min, sec;

  es=self->element_size;
  s=self->ob_size/es;
  ii= i < 0 ? i+s : i;
  if(ii < 0 || ii >= s) return index_error(i);

  switch(self->element_type)
    {
    case STRING_CASES:
      return PyString_FromStringAndSize(((char*)(self->val.CHAR))+ii*es, es);

#define INT_ITEM_CASE(T)  case CS_ ## T ## _TYPE: return PyInt_FromLong((long)(self->val.T[ii]))
    
    INT_ITEM_CASE(BIT);
    INT_ITEM_CASE(TINYINT);
    INT_ITEM_CASE(SMALLINT);
    INT_ITEM_CASE(INT);

    case CS_DATETIME_TYPE:
      ymd(self->val.DATETIME[ii].dtdays, &year, &month, &day);
      sec=self->val.DATETIME[ii].dttime;
      hour=sec/1080000; /* sec / (3600*300) */
      sec %= 1080000; /* sec %= (3600*300) */
      min=sec/18000; /* sec / (60*300) */
      sec %= 18000; /* sec %= (60*300) */
      return Py_BuildValue("iiiiid", year, month, day, hour, min, 
			   ((double)sec)/300.0);

    case CS_DATETIME4_TYPE:
      ymd(self->val.DATETIME4[ii].days, &year, &month, &day);
      min=self->val.DATETIME4[ii].minutes;
      hour=min/60;
      min %= 60;
      return Py_BuildValue("iiiii", year, month, day, hour, min);

    case CS_NUMERIC_TYPE:
    case CS_DECIMAL_TYPE:
      return Py_BuildValue("iis#", 
			   (int)(self->val.NUMERIC[ii].precision),
			   (int)(self->val.NUMERIC[ii].scale),
			   (self->val.NUMERIC[ii].array), CS_MAX_NUMLEN
			   );
    case CS_REAL_TYPE:
      return PyFloat_FromDouble((double)(self->val.REAL[ii]));
    case CS_FLOAT_TYPE:
      return PyFloat_FromDouble((double)(self->val.FLOAT[ii]));
    case CS_MONEY_TYPE:
      UNLESS (h=PyLong_FromLong((long)self->val.MONEY[ii].mnyhigh))
	return NULL;
      UNLESS (l=PyInt_FromLong(32)) goto mnyerr;
      ASSIGN (h, PyNumber_Lshift(h, l));
      UNLESS (h) goto mnyerr;

      ASSIGN (l, PyLong_FromUnsignedLong((unsigned long)
					 self->val.MONEY[ii].mnylow));
      UNLESS(l) goto mnyerr;
      
      ASSIGN(h, PyNumber_Add(h,l));
      
      Py_DECREF(l);
      
      return h;

    case CS_MONEY4_TYPE:
      return PyInt_FromLong((long)(self->val.MONEY4[ii].mny4));
    }

  /* We should never get here */
  Py_INCREF(Py_None);
  return Py_None;

mnyerr:
  Py_XDECREF(h);
  Py_XDECREF(l);
  return NULL;
}

static PyObject *
Buffer_slice(Buffer *self, int ilow, int ihigh)
{
  int il, ih;

  switch (self->element_type)
    {
    case STRING_CASES:
      il= ilow < 0 ? ilow+self->ob_size : ilow;
      ih= ihigh < 0 ? ihigh+self->ob_size : ihigh;
      
      if(il < 0) il=0;
      else if(il > self->ob_size) il=self->ob_size;
      if(ih < 0) ih=0;
      else if(ih > self->ob_size) ih=self->ob_size;
      
      return PyString_FromStringAndSize(&(self->val.c[il]),ih-il);
    }

  return SybaseError("Buffer slicing is only supported for character buffers."
		     );
}

static int 
buffer_ass(Buffer *self, int i, int l, PyObject *v)
{
  int vl;
  char *src;

  if(!v)
    {
      SybaseError("deletion of buffer elements is not allowed");
      return -1;
    }

  if((vl=PyObject_Length(v)) < 0) return -1;
  if(vl > l)
    {
      PyErr_SetString(PyExc_TypeError,
		      "assignment would change the size of the buffer");
      return -1;
    }

  if(v->ob_type==&BufferType) src=(((Buffer        *)v)->val.c);
  else if PyString_Check(v)   src=(((PyStringObject*)v)->ob_sval);
  else
    {
      PyErr_SetString(PyExc_TypeError,
		      "buffer assignment is only allowed from buffers\n"
		      "and strings.");
      return -1;
    }

  memcpy(self->val.c+i, src, vl);
  return 0;
}

static int
Buffer_ass_item(Buffer *self, int i, PyObject *v)
{
  int ii, s, es, w1, w2, l;
  char *array;
  long lv;
  double dv;
  int year, month, day, hour=0, min=0;
  double sec=0;

  es=self->element_size;
  s=self->ob_size/es;
  ii= i < 0 ? i+s : i;
  if(ii < 0 || ii >= s) return index_error(i), -1;

  if(!v)
    {
      PyErr_SetString(PyExc_TypeError,
		      "deletion of buffer elements is not allowed");
      return -1;
    }

  switch(self->element_type)
    {
    case STRING_CASES:
      return buffer_ass(self, ii*es, es, v);
    case CS_DATETIME_TYPE:
      UNLESS (PyArg_ParseTuple(v,"iii|iid",
			       &year,&month,&day, &hour, &min, &sec)) 
	return -1;

      self->val.DATETIME[ii].dtdays=(CS_INT)days(year, month, day)+365;
      self->val.DATETIME[ii].dttime=(CS_INT)(hour*3600+min*60+(int)(sec*300));
      return 0;
    case CS_DATETIME4_TYPE:
      UNLESS (PyArg_ParseTuple(v,"iii|ii",
			       &year,&month,&day, &hour, &min)) 
	return -1;

      self->val.DATETIME4[ii].days=(CS_INT)days(year, month, day)+365;
      self->val.DATETIME4[ii].minutes=(CS_INT)(hour*60+min);
      return 0;
    case CS_NUMERIC_TYPE:
      UNLESS (PyArg_ParseTuple(v,"iis#",&w1,&w2,&array,&l)) return -1;
      self->val.NUMERIC[ii].precision=(CS_BYTE)w1;
      self->val.NUMERIC[ii].scale=(CS_BYTE)w2;
      memcpy(self->val.NUMERIC[ii].array, array, 
	     l > CS_MAX_NUMLEN ? CS_MAX_NUMLEN : l);
      return 0;
    case CS_REAL_TYPE:
    case CS_FLOAT_TYPE:
      if(PyFloat_Check(v))
	dv=PyFloat_AsDouble(v);
      else if((v=PyNumber_Float(v)))
	{
	  dv=PyFloat_AsDouble(v);
	  Py_DECREF(v);
	}
      else return -1;

      if (self->element_type = CS_REAL_TYPE) self->val.REAL[ii]=(CS_REAL)dv;
      else                                   self->val.FLOAT[ii]=(CS_FLOAT)dv;

      return 0;
    case CS_MONEY_TYPE:
      UNLESS (PyArg_ParseTuple(v,"ii",&w1,&w2)) return -1;
      self->val.MONEY[ii].mnyhigh=(CS_INT)w1;
      self->val.MONEY[ii].mnylow=(CS_UINT)w2;
      return 0;
    }


  if(PyInt_Check(v))
    lv=PyInt_AsLong(v);
  else if((v=PyNumber_Int(v)))
    {
      lv=PyInt_AsLong(v);
      Py_DECREF(v);
    }
  else return -1;

  switch(self->element_type)
    {  

#define INT_ASS_ITEM_CASE(T) \
    case CS_ ## T ## _TYPE: \
      self->val.T[ii]=(CS_ ## T)lv; \
      return 0;
    
    INT_ASS_ITEM_CASE(BIT);
    INT_ASS_ITEM_CASE(TINYINT);
    INT_ASS_ITEM_CASE(SMALLINT);
    INT_ASS_ITEM_CASE(INT);

    case CS_MONEY4_TYPE:
      self->val.MONEY4[ii].mny4=lv;
    }

  return 0;
}

static int
Buffer_ass_slice(Buffer *self, int ilow, int ihigh, PyObject *v)
{
  int il, ih;

  switch (self->element_type)
    {
    case STRING_CASES:

      il= ilow < 0 ? ilow+self->ob_size : ilow;
      ih= ihigh < 0 ? ihigh+self->ob_size : ihigh;

      if(il < 0) il=0;
      else if(il > self->ob_size) il=self->ob_size;
      if(ih < 0) ih=0;
      else if(ih > self->ob_size) ih=self->ob_size;

      return buffer_ass(self, il, ih-il, v);
    }
  PyErr_SetString(PyExc_TypeError,
		  "Buffer slicing is only supported for character buffers."
		  );
  return -1;
}

static PySequenceMethods Buffer_as_sequence = {
	(inquiry)Buffer_length,		/*sq_length*/
	(binaryfunc)Buffer_concat,		/*sq_concat*/
	(intargfunc)Buffer_repeat,		/*sq_repeat*/
	(intargfunc)Buffer_item,		/*sq_item*/
	(intintargfunc)Buffer_slice,		/*sq_slice*/
	(intobjargproc)Buffer_ass_item,	/*sq_ass_item*/
	(intintobjargproc)Buffer_ass_slice,	/*sq_ass_slice*/
};

static PyObject *
Buffer_GetAttr(Buffer *self, char *name)
{
  switch (*name) 
    {
    case 'n':
      if (strcmp(name,"typecode")==0)
	return PyInt_FromLong(self->element_type);
      break;
    case 'i':
      if (strcmp(name,"itemsize")==0)
	return PyInt_FromLong(self->element_size);
      break;
    }

  PyErr_SetString(PyExc_AttributeError, name);
  return NULL;
}

static PyTypeObject BufferType = {
  PyObject_HEAD_INIT(NULL) 0,
  "Buffer",			 	/*tp_name*/
  sizeof(Buffer)-sizeof(BufferUnion), 	/*tp_basicsize*/
  1,					/*tp_itemsize*/
  (destructor)Buffer_dealloc,		/*tp_dealloc*/
  (printfunc)0,				/*tp_print*/
  (getattrfunc)Buffer_GetAttr,		/*obsolete tp_getattr*/
  (setattrfunc)0,			/*obsolete tp_setattr*/
  (cmpfunc)0,	/*tp_compare*/
  (reprfunc)0,		/*tp_repr*/
  0,		/*tp_as_number*/
  &Buffer_as_sequence,		/*tp_as_sequence*/
  0,		/*tp_as_mapping*/
  (hashfunc)0,		/*tp_hash*/
  (ternaryfunc)0,	/*tp_call*/
  (reprfunc)Buffer_str,		/*tp_str*/
  (getattrofunc)0,	/*tp_getattro*/
  (setattrofunc)0,	/*tp_setattro*/
  
  0L,0L,
  "Mutable fixed-length data buffers"
};

static PyObject *
newBufferobject(PyObject *ignored, PyObject *args)
{
  int size=1;
  CS_INT element_type;
  int element_size=sizeof(CS_CHAR);
  Buffer *self;

  UNLESS(PyArg_ParseTuple(args,"i|ii:Buffer",
			  &element_type,&size,&element_size)) 
    return NULL;
  
  switch(element_type)
    {
    case CS_TINYINT_TYPE: element_size=sizeof(CS_TINYINT); break;
    case CS_SMALLINT_TYPE: element_size=sizeof(CS_SMALLINT); break;
    case CS_INT_TYPE: element_size=sizeof(CS_INT); break;
    case CS_REAL_TYPE: element_size=sizeof(CS_REAL); break;
    case CS_FLOAT_TYPE: element_size=sizeof(CS_FLOAT); break;
    case CS_MONEY_TYPE: element_size=sizeof(CS_MONEY); break;
    case CS_MONEY4_TYPE: element_size=sizeof(CS_MONEY4); break;
    case CS_DATETIME_TYPE: element_size=sizeof(CS_DATETIME); break;
    case CS_DATETIME4_TYPE: element_size=sizeof(CS_DATETIME4); break;
    case CS_NUMERIC_TYPE: element_size=sizeof(CS_NUMERIC); break;
    case CS_BIT_TYPE: element_size=sizeof(CS_BIT); break;
    case STRING_CASES: break;
    default:
      return SybaseError("Invalid buffer element type");
    }

  UNLESS(self = PyObject_NEW_VAR(Buffer, &BufferType,
				 size*element_size)) return NULL;
  self->element_type=element_type;
  self->element_size=element_size;
  memset(self->val.c, 0, size*element_size);
  return OBJECT(self);
}

#define SY_OCCBF_ID "$Id: sy_occbf.c,v 1.2 1999/10/05 13:51:32 jim Exp $\n"
